#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>


#define MAX_ROWS      9
#define MAX_COLUMNS   9


#define positionNone 0
#define positionHole 1
#define positionFull 2


#define POSITIONS_PER_BYTE     4
#define XY(x,y)                (((y)*MAX_COLUMNS)+(x))
#define GETPOSITION(board,x,y) (board[XY(x,y)/POSITIONS_PER_BYTE]&\
                               (3<<(2*((XY(x,y)%POSITIONS_PER_BYTE))))\
                               >>(2*(XY(x,y)%POSITIONS_PER_BYTE)))


/* All numbers are BIG endian */

struct pdbHeader {
  char     name[32];		/* Database name (no NUL termination) */
  uint16_t attrs;		/* Flags */
  uint16_t version;		/* Version number */
  uint32_t ctime;		/* Creation time (since 1/1/1904) */
  uint32_t mtime;		/* Last Modification time (=ctime initially) */
  uint32_t btime;		/* Last Backup time (0 initially) */
  uint32_t modnum;		/* Modification Number (always 0) */
  uint32_t appinfoarea;		/* Pointer to the AppInfoArea block (0=none) */
  uint32_t sortinfoarea;	/* Pointer to the SortInfoArea block (always 0) */
  char     dbtype[4];		/* Database type (no NUL) */
  char     creator[4];		/* CreatorID (no NUL) */
  uint32_t seed;		/* Always 0x00000000 */
  uint32_t next;		/* Always 0x00000000 */
  uint32_t numrecords;		/* Number of records in database */
};

struct pdbRecListItem {
  uint32_t offset;		/* Offset of corresponding data in PDB file */
  char     attrs;		/* Record flags/category */
  char     recordID[3];		/* Always 0x000000 */
};

struct NativeLevel {
  char name[33];		/* 32 bytes + NUL */
  char data[82];		/* 9x9 + NUL */
};


static struct NativeLevel levels []={
    {"Classic",
       "         "
       "   ***   "
       "   ***   "
       " ******* "
       " ***o*** "
       " ******* "
       "   ***   "
       "   ***   "
       "         "},

    {"Fat Classic",
       "  *****  "
       "  *****  "
       "*********"
       "*********"
       "****o****"
       "*********"
       "*********"
       "  *****  "
       "  *****  "},

    {"Ex",
       "***   ***"
       "***   ***"
       "*********"
       "  *****  "
       "  **o**  "
       "  *****  "
       "*********"
       "***   ***"
       "***   ***"},

    {"Diamond",
       "    *    "
       "   ***   "
       "  *****  "
       " ******* "
       "****o****"
       " ******* "
       "  *****  "
       "   ***   "
       "    *    "},

    {"Lift Buttons",
       "    *    "
       "   ***   "
       "  *****  "
       " ******* "
       "    o    "
       " ******* "
       "  *****  "
       "   ***   "
       "    *    "},

    {"Hourglass",
       "*********"
       " ******* "
       "  *****  "
       "   ***   "
       "    o    "
       "   ***   "
       "  *****  "
       " ******* "
       "*********"},

    {"Surreal",
       "***** ***"
       "***** ***"
       "*****o***"
       "****    o"
       "****o****"
       "o    ****"
       "***o*****"
       "*** *****"
       "*** *****"},

    {"",""}
};


int main()
{
  struct pdbHeader hdr;
  unsigned char binboard[256];
  int j;

  bzero(&hdr,sizeof(hdr));
  strncpy(hdr.name,"SOLO Levels",strlen("SOLO Levels"));
  

#if 0
struct pdbHeader {
  char     name[32];		/* Database name (no NUL termination) */
  uint16_t attrs;		/* Flags */
  uint16_t version;		/* Version number */
  uint32_t ctime;		/* Creation time (since 1/1/1904) */
  uint32_t mtime;		/* Last Modification time (=ctime initially) */
  uint32_t btime;		/* Last Backup time (0 initially) */
  uint32_t modnum;		/* Modification Number (always 0) */
  uint32_t appinfoarea;		/* Pointer to the AppInfoArea block (0=none) */
  uint32_t sortinfoarea;	/* Pointer to the SortInfoArea block (always 0) */
  char     dbtype[4];		/* Database type (no NUL) */
  char     creator[4];		/* CreatorID (no NUL) */
  uint32_t seed;		/* Always 0x00000000 */
  uint32_t next;		/* Always 0x00000000 */
  uint32_t numrecords;		/* Number of records in database */
};
#endif

  for(j=0;levels[j].name[0];j++){
    int  x,y,shift=0,i=0,num_pieces=0;
    char *board=levels[j].data;

    memset(binboard,0,sizeof(binboard));
    printf("{\"%s\",\n\t{",levels[j].name);
    for(y=0;y<9;y++)for(x=0;x<9;x++){
      int val;
      switch(board[y*9+x]){
      case '*': val=2; num_pieces++; break;
      case 'o': val=1; break;
      default:  val=0; break;
      }
      
      binboard[i]|=(val<<shift);
      shift+=2;
      if(shift>7){
	printf("0x%02x,",binboard[i]);
	i++;
	if((i%8)==0)printf("\n\t");
	shift=0;
      }
    }
    if(shift){
      printf("0x%02x},\n\t%d},\n",binboard[i],num_pieces);
    }
    
#if 0
    for(y=0;y<9;y++){
      printf("\n");
      for(x=0;x<9;x++){
	printf("%02x ",((binboard[(y*MAX_COLUMNS+x)/4])&
			(3<<(2*((y*MAX_COLUMNS+x)%4))))
	       >>(2*((y*MAX_COLUMNS+x)%4)));
      }
    }
#endif
  }

  return 0;
}
