#include <Pilot.h>
#include "solo.h"

typedef unsigned char PositionType;

#define soloAppVersionNum 0x0001
#define soloAppCreator    CREATOR /* Makefile defines this */

#define soloDBName    "SoloLevels"
#define soloDBType    'DATA'

#define PIECE_HEIGHT 16
#define PIECE_WIDTH  16
#define NUM_BITMAPS   2


#define MAX_ROWS      9
#define MAX_COLUMNS   9


#define positionNone 0
#define positionHole 1
#define positionFull 2


#define POSITIONS_PER_BYTE       4


struct Level {
  char name[32];		/* Level name, not null terminated */
  char data[21];		/* 9x9x2 bits = 21 bytes of level data */
  char count;			/* Initial number of pieces (0--81) */
};



#ifdef GENERATE_FULL_PDB
#define MAX_LEVEL 7
#else
#define MAX_LEVEL 1
#endif


struct Level prototype_levels[MAX_LEVEL] = {
  {"Classic",
   {0x00,0x00,0x00,0x2a,0x00,0xa8,0x00,0xaa,
    0x2a,0xa8,0xa9,0xa0,0xaa,0x02,0xa8,0x00,
    0xa0,0x02,0x00,0x00,0x00},
   32},
#ifdef GENERATE_FULL_PDB
  {"Fat Classic",
   {0xa0,0x2a,0x80,0xaa,0xa0,0xaa,0xaa,0xaa,
    0xaa,0xaa,0xa9,0xaa,0xaa,0xaa,0xaa,0x2a,
    0xa8,0x0a,0xa0,0x2a,0x00},
   64},
  {"Ex",
   {0x2a,0xa0,0xaa,0x80,0xaa,0xaa,0x2a,0xa8,
    0x0a,0xa0,0x29,0x80,0xaa,0xa0,0xaa,0xaa,
    0x0a,0xa8,0x2a,0xa0,0x02},
   56},
  {"Diamond",
   {0x00,0x02,0x00,0x2a,0x00,0xaa,0x02,0xaa,
    0x2a,0xaa,0xa9,0xa2,0xaa,0x02,0xaa,0x02,
    0xa0,0x02,0x00,0x02,0x00},
   40},
  {"Lift Buttons",
   {0x00,0x02,0x00,0x2a,0x00,0xaa,0x02,0xaa,
    0x2a,0x00,0x01,0xa0,0xaa,0x02,0xaa,0x02,
    0xa0,0x02,0x00,0x02,0x00},
   32},
  {"Hourglass",
   {0xaa,0xaa,0xa2,0xaa,0x02,0xaa,0x02,0xa0,
    0x02,0x00,0x01,0x00,0x2a,0x00,0xaa,0x02,
    0xaa,0x2a,0xaa,0xaa,0x02},
   48},
  {"Surreal",
   {0xaa,0xa2,0xaa,0x8a,0xaa,0x6a,0xaa,0x2a,
    0x40,0xaa,0xa9,0x06,0xa0,0xaa,0xa6,0xaa,
    0x8a,0xaa,0x2a,0xaa,0x02},
   64},
#endif
};




#if 0
const static char num_pieces=32;
#endif
const static char num_pieces=64;

const static PositionType new_board[]={
#if 0
  0x00,0x00,0x00,0x2a,0x00,0xa8,0x00,0xaa,
  0x2a,0xa8,0xa9,0xa0,0xaa,0x02,0xa8,0x00,
  0xa0,0x02,0x00,0x00,0x00
#endif
  0xaa,0xa2,0xaa,0x8a,0xaa,0x6a,0xaa,0x2a,
  0x40,0xaa,0xa9,0x06,0xa0,0xaa,0xa6,0xaa,
  0x8a,0xaa,0x2a,0xaa,0x02
};


typedef struct SoloState {
  Boolean        restored;
  PointType      origin;
  char           px;
  char           py;
  char           num_pieces;
  char           in_progress;
  PositionType	 board[MAX_ROWS*((MAX_COLUMNS+POSITIONS_PER_BYTE-1)/POSITIONS_PER_BYTE)];
} SoloState;



typedef struct SoloPrefType {
  DWord	    magic;
  int       level;
  SoloState game;
} SoloPrefType;


/*
 *  This spells "SOLO". MSW is in hex (0x53='S'), LSW in dec (76='L', 79='O').
 */

#define SOLO_MAGIC 0x534F7679


static DmOpenRef     SoloDB;

static WinHandle     OffscreenBitmapWinH=0;

static SoloPrefType  SoloPref;

#define Game SoloPref.game




static PositionType
get_position(PositionType *board, char x, char y)
{
  register SWord point  = y*MAX_COLUMNS + x;
  register SWord offset = point / POSITIONS_PER_BYTE;
  register SWord mask   = 2 * (point % POSITIONS_PER_BYTE);
  
  return (board [offset] & (3 << mask)) >> mask;
}


static void
set_position(PositionType *board, char x, char y, char v)
{
  register SWord point  = y*MAX_COLUMNS + x;
  register SWord offset = point / POSITIONS_PER_BYTE;
  register SWord mask   = 2 * (point % POSITIONS_PER_BYTE);
  
  board [offset] &= ~ (3 << mask);
  board [offset] |= v << mask;
}


static int
FindMoreMoves()
{
  int x,y;
 
  for (y=0;y<MAX_ROWS;y++)for(x=0;x<MAX_COLUMNS;x++){
    /* Check for a full position */
    if(get_position(Game.board,x,y)!=positionFull)continue;

    /* Check for a move to the left */
    if(x>1){
      if((get_position(Game.board,x-1,y)==positionFull) &&
	 (get_position(Game.board,x-2,y)==positionHole)){
	return 1;
      }
    }

    /* Check for a move to the right */
    if(x<MAX_COLUMNS-2){
      if((get_position(Game.board,x+1,y)==positionFull) &&
	 (get_position(Game.board,x+2,y)==positionHole)){
	return 1;
      }
    }

    /* Check for a move up */
    if(y>1){
      if((get_position(Game.board,x,y-1)==positionFull) &&
	 (get_position(Game.board,x,y-2)==positionHole)){
	return 1;
      }
    }

    /* Check for a move down */
    if(y<MAX_ROWS-2){
      if((get_position(Game.board,x,y+1)==positionFull) &&
	 (get_position(Game.board,x,y+2)==positionHole)){
	return 1;
      }
    }
  }

  return 0;
}


static void
DisplayNumPieces()
{
  char s[10];
  RectangleType rect;

  rect.topLeft.x=80;
  rect.topLeft.y=1;
  rect.extent.x=20;
  rect.extent.y=10;
  WinEraseRectangle(&rect,1);
  FntSetFont(1);
  StrPrintF(s,"%d",Game.num_pieces);
  WinDrawChars(s,StrLen(s),80,1);
}


static void
DisplayBoard()
{
  int            x,y;
  RectangleType  rect,er;
 
  rect.extent.x=er.extent.x=PIECE_WIDTH;
  rect.extent.y=er.extent.y=PIECE_HEIGHT;
  rect.topLeft.x=0;
  er.topLeft.y=16;

  Game.num_pieces=0;
  for (y=0;y<MAX_ROWS;y++,er.topLeft.y+=PIECE_HEIGHT){

    er.topLeft.x=8;
    for(x=0;x<MAX_COLUMNS;x++,er.topLeft.x+=PIECE_WIDTH){

      switch(get_position(Game.board,x,y)){
      case positionHole:
	rect.topLeft.y=0;
	WinCopyRectangle(OffscreenBitmapWinH,WinGetDisplayWindow(),&rect,
			 er.topLeft.x,er.topLeft.y,scrCopy);
	break;

      case positionFull:
	Game.num_pieces++;
	rect.topLeft.y=PIECE_WIDTH;
	WinCopyRectangle(OffscreenBitmapWinH,WinGetDisplayWindow(),&rect,
			 er.topLeft.x,er.topLeft.y,scrCopy);
	break;

      default:
      case positionNone:
	WinEraseRectangle(&er,0);
      }
    }
  }

  DisplayNumPieces();
}



static void
NewBoard()
{
  VoidHand levelRec=DmGetRecord(SoloDB,SoloPref.level);
  struct Level *level;

  ErrFatalDisplayIf(levelRec==0,"board not found");

  level=(struct Level*)MemHandleLock(levelRec);
  MemMove(Game.board,&(level->data),sizeof(Game.board));
  MemPtrUnlock(level);
  DmReleaseRecord(SoloDB,SoloPref.level,true);

  Game.num_pieces=num_pieces;
  Game.px=Game.py=-1;
  Game.in_progress=1;
}



static void
LoadBoard()
{
  Boolean found;
  
  found=PrefGetAppPreferencesV10(soloAppCreator,soloAppVersionNum,&SoloPref,sizeof(SoloPref));
  if (found &&
      SoloPref.magic==SOLO_MAGIC &&
      Game.in_progress){
  } else {
    SoloPref.magic=SOLO_MAGIC;
    SoloPref.level=0;
    NewBoard();
  }
}


static void
SaveBoard(void)
{
  SoloPref.magic=SOLO_MAGIC;
  PrefSetAppPreferencesV10(soloAppCreator,soloAppVersionNum,
			   &SoloPref,sizeof(SoloPref));
}


static void
MakeBitmaps()
{
  Word error;

  if (!OffscreenBitmapWinH){
    Int	i;
    OffscreenBitmapWinH=
      WinCreateOffscreenWindow (PIECE_WIDTH, PIECE_HEIGHT*NUM_BITMAPS,
				screenFormat, &error);

    if(OffscreenBitmapWinH) {
      Handle    resH;
      BitmapPtr resP;
      WinHandle currDrawWindow=WinSetDrawWindow(OffscreenBitmapWinH);
      
      for (i=0;i<NUM_BITMAPS;i++) {
	resH=(Handle)DmGetResource(bitmapRsc,bitmapID_hole+i);
	ErrFatalDisplayIf(!resH,"no bitmap");
	resP=(BitmapPtr)MemHandleLock((VoidHand)resH);
	WinDrawBitmap(resP,0,PIECE_HEIGHT*i);
	MemPtrUnlock(resP);
	DmReleaseResource((VoidHand)resH);
      }
      
      WinSetDrawWindow(currDrawWindow);
    } else {
      ErrDisplay("unallocated bitmap window");
    }
  }
}

static void
make_move(SWord x, SWord y)
{
  int px,py;
  RectangleType rect;

  if(x<8||y<16)return;
  px=(x-8)/PIECE_WIDTH;
  py=(y-16)/PIECE_HEIGHT;
  rect.extent.x=15;
  rect.extent.y=15;

  if(px<0 || px>=MAX_COLUMNS || py<0 || py>=MAX_COLUMNS)return;

  switch(get_position(Game.board,px,py)){
  case positionFull:
    SndPlaySystemSound(sndClick);

    /* Player changed their mind. We might have to 'uninvert' the previous
       selection. */

    if(Game.px>=0){
      rect.topLeft.x=8+Game.px*PIECE_WIDTH;
      rect.topLeft.y=16+Game.py*PIECE_HEIGHT;
      WinInvertRectangle(&rect,4);
    }
    rect.topLeft.x=8+px*PIECE_WIDTH;
    rect.topLeft.y=16+py*PIECE_HEIGHT;
    WinInvertRectangle(&rect,4);
    Game.px=px;
    Game.py=py;
    break;
    
  case positionHole:

    /* Can't move a hole! */
    
    if(Game.px<0){
      SndPlaySystemSound(sndWarning);
      break;
    }

    /* Make sure we're moving orthogonally by exactly 2 places */

    if(!(((abs(px-Game.px)==2) && (py==Game.py)) ||
	 ((abs(py-Game.py)==2) && (px==Game.px)))){
      SndPlaySystemSound(sndWarning);
      break;
    }

    /* We MUST jump over an existing piece! */
    if(get_position(Game.board,(Game.px+px)>>1,(Game.py+py)>>1)!=positionFull){
      SndPlaySystemSound(sndWarning);
      break;
    }

    /* Move, removing the piece we jumped over */

    SndPlaySystemSound(sndClick);
    set_position(Game.board,px,py,positionFull);
    set_position(Game.board,Game.px,Game.py,positionHole);
    set_position(Game.board,(Game.px+px)>>1,(Game.py+py)>>1,positionHole);

    Game.px=-1;
    DisplayBoard();
    if(!FindMoreMoves()){
      int alertID;
      Game.in_progress=0;
      switch(Game.num_pieces){
      case 1:
	alertID=GameWonAlert;
	break;
      case 2:
      case 3:
	alertID=NotBadAlert;
	break;
      case 4:
      case 5:
      case 6:
	alertID=CanDoBetterAlert;
	break;
      default:
	alertID=NoMoreMovesAlert;
      }
      if(FrmAlert(alertID)){
	NewBoard();
	DisplayBoard();
      }
    }
    break;

  default:
    //SndPlaySystemSound(sndWarning);

    /* Player changed their mind. We might have to 'uninvert' the previous
       selection. */

    if(Game.px>=0){
      rect.topLeft.x=8+Game.px*PIECE_WIDTH;
      rect.topLeft.y=16+Game.py*PIECE_HEIGHT;
      WinInvertRectangle(&rect,4);
    }
    Game.px=-1;
  }
}


void
Levels()
{
  FormPtr        frm;
  int            i,j;
  CharPtr       *opts, label;
  VoidPtr        p;
  struct Level  *level;
  /*DmOpenRef      SoloDB=DmOpenDatabaseByTypeCreator(soloDBType, CREATOR, dmModeReadWrite);*/

  j=DmNumRecords(SoloDB);
  /*j=16;*/
  if(j==0)return;		/* Famous Last Words: This Should Never Happen */

  opts=(CharPtr*)MemPtrNew(j*sizeof(char*));
  if(opts==0)return;

  for(i=0;i<j;i++){
    VoidHand levelRec=DmGetRecord(SoloDB,i);
    
    level=(struct Level*)MemHandleLock(levelRec);
    opts[i]=MemPtrNew(sizeof(char)*(StrLen(level->name)+1));
    StrCopy(opts[i],level->name);
    MemPtrUnlock (level);
    DmReleaseRecord (SoloDB,i,true);
  }

  frm=FrmInitForm(LevelsForm);
  p=FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,listID_levels));
  LstSetListChoices(p,opts,j);
  LstSetSelection(p,SoloPref.level);
  label=LstGetSelectionText(p,SoloPref.level);
  CtlSetLabel(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,popuptriggerID_levels)),label);
  
  j=0;
  if(FrmDoDialog(frm)==LevelsOKButton){
    SoloPref.level=LstGetSelection(p);
    NewBoard();
    j=1;
  }
  FrmDeleteForm(frm);

  for(i=0;i<j;i++) MemPtrFree(opts[i]);
  MemPtrFree(opts);

  if(j)DisplayBoard();
}


static Boolean
MainFormHandleEvent(EventPtr event)
{
  Boolean handled=false;
  FormPtr frm;
  static char armed=0;
  
  switch (event->eType) {
  case frmOpenEvent:
    frm = FrmGetActiveForm();
    FrmDrawForm(frm);
    DisplayBoard();
    handled = true;
    break;
    
  case ctlSelectEvent:
    if (event->data.ctlEnter.controlID==buttonID_newgame) {
      NewBoard();
      DisplayBoard();
    }
    break;

  case penDownEvent:
    armed=1;
    break;
  case penUpEvent:
    if(armed){
      make_move(event->data.penUp.end.x,event->data.penUp.end.y);
      armed=0;
    }
    break;
    
  case menuEvent:
    MenuEraseStatus(MenuGetActiveMenu());
    switch(event->data.menu.itemID){
    case MainGameLevel:
      Levels();
      handled = true;
      break;
    case MainGameAbout:
      frm = FrmInitForm(AboutForm);
      FrmDoDialog(frm);
      FrmDeleteForm(frm);
      handled = true;
      break;
    case MainGameInstr:
      FrmHelp(stringID_Instructions);
      handled = true;
      break;
    }
    break;

  default:
    break;			/* To stop spurious warning */
  }
  return handled;
}


static Boolean
ApplicationHandleEvent(EventPtr event)
{
  FormPtr  frm;
  Int      formId;
  Boolean  handled=false;

  if (event->eType == frmLoadEvent) {
    formId=event->data.frmLoad.formID;
    frm=FrmInitForm(formId);
    FrmSetActiveForm(frm);
    
    switch(formId) {
    case MainForm:
      FrmSetEventHandler(frm,MainFormHandleEvent);
      break;
    }
    handled = true;
  }
  return handled;
}


static void
EventLoop()
{
  EventType event;
  Word      error;

  do {
    EvtGetEvent(&event, 2);
    if (!SysHandleEvent(&event)){
      if (!MenuHandleEvent(0, &event, &error)){
	if (!ApplicationHandleEvent(&event)){
	  FrmDispatchEvent(&event);
	}
      }
    }
  } while (event.eType!=appStopEvent);
}


static void
OpenOrCreatePDB()
{
  UInt      cardNo;
  LocalID   dbID;
  Err       err;
  Word      attributes, i;

  SoloDB=DmOpenDatabaseByTypeCreator(soloDBType, CREATOR, dmModeReadWrite);
  if(!SoloDB){
    err=DmCreateDatabase (0, soloDBName, CREATOR, soloDBType, false);
    if(err)return;
    
    SoloDB = DmOpenDatabaseByTypeCreator(soloDBType, CREATOR, dmModeReadWrite);
    if(!SoloDB) return;

    // Set the backup bit.  This is to aid syncs with non Palm software.
    DmOpenDatabaseInfo(SoloDB, &dbID, NULL, NULL, &cardNo, NULL);
    DmDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
		   NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    attributes |= dmHdrAttrBackup;
    DmSetDatabaseInfo(cardNo, dbID, NULL, &attributes, NULL, NULL,
		      NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    
    //err = SoloAppInfoInit (SoloDB);
    if (err) {
      DmCloseDatabase(SoloDB);
      DmDeleteDatabase(cardNo, dbID);
      return;
    }

    for(i=0;i<MAX_LEVEL;i++){
      VoidHand levelRec = DmNewRecord (SoloDB, &i, sizeof(struct Level));
      Ptr p;
      if (!levelRec) {
	FrmAlert (DeviceFullAlert);
	return;
      }
      p=MemHandleLock(levelRec);
      DmWrite (p,0,&prototype_levels[i],sizeof(struct Level));
      MemPtrUnlock (p);
      DmReleaseRecord (SoloDB,i,true);
    }
  }
}



static Boolean
StartApplication()
{
  OpenOrCreatePDB();

  FrmGotoForm(MainForm);
  MakeBitmaps();
  LoadBoard();

  return false;
}


static void
StopApplication()
{
  FrmCloseAllForms();
  SaveBoard();
  if (OffscreenBitmapWinH){
    WinDeleteWindow(OffscreenBitmapWinH,false);
    OffscreenBitmapWinH=0;
  }
  DmCloseDatabase(SoloDB); 
}


DWord
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{

  if(cmd==sysAppLaunchCmdNormalLaunch) {
    if (!StartApplication()) {
      EventLoop();
      StopApplication();
    }
  }
  return 0;
}
